Castbars = LibStub("AceAddon-3.0"):NewAddon("Castbars");

Castbars.SharedMedia = LibStub("LibSharedMedia-3.0");

function Castbars:FrameLayoutRestore(frames)
    for i, frame in pairs(frames) do
        local frameName = frame:GetName();

        -- Position
        local position = self.db.profile[frameName .. "Position"];
        frame:ClearAllPoints();
        frame:SetPoint(position.point, position.parent, position.relpoint, position.x, position.y);

        -- Texture
        local barTexture = self.SharedMedia:Fetch('statusbar', self.db.profile[frameName .. "Texture"]);
        local frameBackdrop = _G[frameName .. "Backdrop"];
        local borderTexture = self.SharedMedia:Fetch('border', self.db.profile[frameName .. "Border"]);
        frame:SetStatusBarTexture(barTexture);
        frameBackdrop:SetBackdrop({edgeFile = borderTexture, edgeSize = 16, tileSize = 16, insets = {bottom = 5, top = 5, right = 5, left = 5}});
        local brightness = self.db.profile[frameName .. "BorderBrightness"];
        frameBackdrop:SetBackdropBorderColor(brightness, brightness, brightness, 1.0);

        -- Dimensions
        frame:SetWidth(self.db.profile[frameName .. "Width"]);
        frame:SetHeight(self.db.profile[frameName .. "Height"]);

        -- Icon position
        local frameIcon = _G[frameName .. "Icon"];
        frameIcon:ClearAllPoints();
        if (self.db.profile[frameName .. "Border"] == "None") then
            frameIcon:SetPoint("RIGHT", frame, "LEFT", 0, 0);
            frameIcon:SetHeight(self.db.profile[frameName .. "Height"]);
            frameIcon:SetWidth(self.db.profile[frameName .. "Height"]);
        else
            frameIcon:SetPoint("RIGHT", frame, "LEFT", -5, 0);
            frameIcon:SetHeight(self.db.profile[frameName .. "Height"] + 5);
            frameIcon:SetWidth(self.db.profile[frameName .. "Height"] + 5);
        end

        -- Visibility
        if (self.db.profile[frameName .. "Show"] == false) then
            frame.showCastbar = false;
        else
            frame.showCastbar = true;
        end
    end
end

function Castbars:FrameCustomize(frames)
    for i, frame in pairs(frames) do
        local frameName = frame:GetName();

        -- Make dragable
        frame:EnableMouse(true);
        frame:SetMovable(true);
        frame:RegisterForDrag("LeftButton");
        frame:SetScript("OnDragStart", function(frame)
            if (self.ConfigMode) then
                frame:StartMoving();
            end
        end);
        frame:SetScript("OnDragStop", function(frame)
            frame:StopMovingOrSizing();
            local position = self.db.profile[frameName .. "Position"];
            position.point, position.parent, position.relpoint, position.x, position.y = frame:GetPoint();
        end);

        -- Adjust spark position
        local frameSpark = _G[frameName .. "Spark"];
        local setPoint = frameSpark.SetPoint;
        frameSpark.SetPoint = function(self, point, relativeFrame, realativePoint, x, y)
            setPoint(self, point, relativeFrame, realativePoint, x, 0);
        end

        -- Adjust text position
        local frameText = _G[frameName .. "Text"];
        frameText:ClearAllPoints();
        frameText:SetPoint("CENTER", frame, "CENTER", 0, 0);

        -- Remove the border
        local frameBorder = _G[frameName .. "Border"];
        frameBorder:SetTexture();

        -- Remove the flash
        local frameFlash = _G[frameName .. "Flash"];
        frameFlash:SetTexture();

        -- Create backdrop
        frame.backdrop = CreateFrame("Frame", frameName .. "Backdrop", frame);
        frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 0);

        -- Make icon visible
        local frameIcon = _G[frameName .. "Icon"];
        frameIcon:Show();

        -- Add timer text
        frame.timer = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlight");
        frame.timer:SetPoint("LEFT", frame, "RIGHT", 5, 0);
        frame.nextupdate = 0.1;

        -- Add latency indication texture
        frame.latency = frame:CreateTexture(nil, "BACKGROUND");
        frame.latency:SetHeight(frame:GetHeight());
        frame.latency:SetPoint("RIGHT", frame, "RIGHT", 0, 0);
        frame.latency:SetTexture(1, 0, 0, 1);

        -- Automatically adjust the height of sub elements
        frame.SetHeightOriginal = frame.SetHeight;
        frame.SetHeight = function(self, height)
            frame.backdrop:SetHeight(height + 10);
            frameSpark:SetHeight(2.5 * height);
            frame.latency:SetHeight(height);
            self:SetHeightOriginal(height);
        end

        -- Automatically adjust the width of sub elements
        frame.SetWidthOriginal = frame.SetWidth
        frame.SetWidth = function(self, width)
            frame.backdrop:SetWidth(width + 10);
            self:SetWidthOriginal(width);
        end

        frame:Hide();
    end
end

function Castbars:GetOptionsTableForBar(frameName, friendlyName, order)
    local optionstable =
    {
        type = "group",
        name = friendlyName .. " Castbar Options",
        guiInline = true,
        order = order,
        args =
        {
            enable =
            {
                name = "Enable",
                type = "toggle",
                order = 1,
                desc = "Toggles display of the " .. friendlyName .. " Castbar",
                get = function() return self.db.profile[frameName .. "Show"] end,
                set = function()
                    self.db.profile[frameName .. "Show"] = not self.db.profile[frameName .. "Show"];
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            width =
            {
                name = "Width",
                type = "range",
                order = 2,
                desc = "Set the width of the " .. friendlyName .. " Castbar",
                min = 100,
                max = 300,
                get = function() return self.db.profile[frameName .. "Width"] end,
                set = function(info, value)
                    self.db.profile[frameName .. "Width"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            height =
            {
                name = "Height",
                type = "range",
                order = 3,
                desc = "Set the height of the " .. friendlyName .. " Castbar",
                min = 10,
                max = 50,
                get = function() return self.db.profile[frameName .. "Height"] end,
                set = function(info, value)
                    self.db.profile[frameName .. "Height"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            texture =
            {
                name = "Texture",
                type = 'select',
                order = 4,
                desc = "Select texture to use for the " .. friendlyName .. " Castbar",
                dialogControl = 'LSM30_Statusbar',
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return self.db.profile[frameName .. "Texture"] end,
                set = function(info, value)
                    self.db.profile[frameName .. "Texture"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            border =
            {
                name = "Border",
                type = 'select',
                order = 5,
                desc = "Select border to use for the " .. friendlyName .. " Castbar",
                dialogControl = 'LSM30_Border',
                values = AceGUIWidgetLSMlists.border,
                get = function() return self.db.profile[frameName .. "Border"] end,
                set = function(info, value)
                    self.db.profile[frameName .. "Border"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            borderbrightness =
            {
                name = "Border Brightness",
                type = "range",
                order = 6,
                desc = "Set the brightness of the border of the " .. friendlyName .. " Castbar",
                min = 0.0,
                max = 1.0,
                isPercent = true,
                get = function() return self.db.profile[frameName .. "BorderBrightness"] end,
                set = function(info, value)
                    self.db.profile[frameName .. "BorderBrightness"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
        },
    };
    return optionstable;
end

function Castbars:GetOptionsTable()
    local optionstable =
    {
        type = "group",
        icon = "",
        name = "Castbars",
        handler = self,
        childGroups = "tree",
        args =
        {
            configmode =
            {
                type = "group",
                name = "Configuration Mode",
                guiInline = true,
                order = 0,
                args =
                {
                    toggleconfigmode =
                    {
                        name = "Toggle",
                        type = "execute",
                        order = 1,
                        desc = "Toggle configuration mode to allow moving bars and setting appearance options",
                        func = "Toggle",
                    },
                },
            },
            player = self:GetOptionsTableForBar("CastingBarFrame", "Player", 1),
            target = self:GetOptionsTableForBar("CastingBarTargetFrame", "Target", 2),
        },
    };
    return optionstable;
end

function Castbars:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("CastbarsDB",
    {
        profile =
        {
            CastingBarFrameShow = true,
            CastingBarFrameWidth = 195,
            CastingBarFrameHeight = 13,
            CastingBarFrameTexture = "Blizzard",
            CastingBarFrameBorder = "Blizzard Dialog",
            CastingBarFrameBorderBrightness = 0.5,
            CastingBarFramePosition = {point = "BOTTOM", relpoint = "BOTTOM", x = 0, y = 200},

            CastingBarTargetFrameShow = true,
            CastingBarTargetFrameWidth = 195,
            CastingBarTargetFrameHeight = 13,
            CastingBarTargetFrameTexture = "Blizzard",
            CastingBarTargetFrameBorder = "None",
            CastingBarTargetFrameBorderBrightness = 0.5,
            CastingBarTargetFramePosition = {point = "CENTER", relpoint = "CENTER", x = 0, y = 200},
        }
    });

    LibStub("AceConfig-3.0"):RegisterOptionsTable("Castbars", self:GetOptionsTable());
    self.BlizzardOptionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", "Castbars");

    -- Create target casting bar
    CreateFrame("StatusBar", "CastingBarTargetFrame", UIParent, "CastingBarFrameTemplate");
    CastingBarTargetFrame:SetScript("OnEvent", function(frame, event, ...)
        if (event == "PLAYER_ENTERING_WORLD") then
            self:FrameLayoutRestore(self.frames);
        elseif (event == "PLAYER_TARGET_CHANGED") then
            event = "PLAYER_ENTERING_WORLD"
        end
        CastingBarFrame_OnEvent(frame, event, ...);
    end)
    CastingBarTargetFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
    CastingBarFrame_OnLoad(CastingBarTargetFrame, "target", false);

    self.frames = {CastingBarFrame, CastingBarTargetFrame};

    -- Customize the bars
    self:FrameCustomize(self.frames);

    -- Add timer handling code
    hooksecurefunc("CastingBarFrame_OnUpdate", function(self, elapsed)
        if (self.timer) then
            if (self.nextupdate < elapsed) then
                if (self.casting) then
                    self.timer:SetText(format("%.1f", max(self.maxValue - self.value, 0)));
                elseif (self.channeling) then
                    self.timer:SetText(format("%.1f", max(self.value, 0)));
                else
                    self.timer:SetText("");
                end;
                self.nextupdate = 0.1;
            else
                self.nextupdate = self.nextupdate - elapsed;
            end;
        end;
    end);

    -- Add latency indication handling code
    hooksecurefunc("CastingBarFrame_OnShow", function(self)
        if (self.latency) then
            local min, max = self:GetMinMaxValues();
            local down, up, lag = GetNetStats();
            local latency = (lag / 1000) / (max - min);
            if (latency < 0) then latency = 0 elseif (latency > 1) then latency = 1 end
            if (self.channeling) then
                self.latency:ClearAllPoints();
                self.latency:SetPoint("LEFT", self, "LEFT", 0, 0);
            else
                self.latency:ClearAllPoints();
                self.latency:SetPoint("RIGHT", self, "RIGHT", 0, 0);
            end;
            if (max > 4) then
                self.latency:Hide();
            else
                self.latency:SetWidth(self:GetWidth() * latency);
                self.latency:Show();
            end
        end
    end);

    self.GetOptionsTableForBar = nil;
    self.GetOptionsTable = nil;
    self.FrameCustomize = nil;
    self.OnInitialize = nil;
end

--[[ ConfigMode Support ]]--

Castbars.CastingBarFrame_OnEvent = CastingBarFrame_OnEvent;
Castbars.DoNothing = function() end;

function Castbars:Show()
    self.ConfigMode = true;
    CastingBarFrame_OnEvent = self.DoNothing;
    CastingBarFrameText:SetText("Player Castbar")
    CastingBarFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    CastingBarFrame:SetAlpha(1)
    CastingBarFrameSpark:Hide();
    CastingBarFrameFlash:Hide();
    CastingBarFrame.fadeOut = nil;
    CastingBarFrame:Show();
    CastingBarTargetFrameText:SetText("Target Castbar")
    CastingBarTargetFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    CastingBarTargetFrame:SetAlpha(1)
    CastingBarTargetFrameSpark:Hide();
    CastingBarTargetFrameFlash:Hide();
    CastingBarTargetFrame.fadeOut = nil;
    CastingBarTargetFrame:Show();
end

function Castbars:Hide()
    CastingBarFrame:Hide();
    CastingBarTargetFrame:Hide();
    CastingBarFrame_OnEvent = self.CastingBarFrame_OnEvent;
    self.ConfigMode = false;
end

function Castbars:Toggle()
    if (self.ConfigMode) then
        self:Hide();
    else
        self:Show();
    end
end

CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {};
CONFIGMODE_CALLBACKS["Castbars"] = function(action)
    if (action == "ON") then Castbars:Show()
    elseif (action == "OFF") then Castbars:Hide() end
end
